"""Observer framework to trigger methods"""

class Observable(object):
    """Observerable object that manages observer methods"""
    def __init__(self):
        self.obs = []
        self.changed = 0

    def add_observer(self, observer):
        """add obserser method if not already connected"""
        if observer not in self.obs:
            self.obs.append(observer)

    def delete_observer(self, observer):
        """remove observer method"""
        self.obs.remove(observer)

    def notify_observers(self, arg=None):
        '''If 'changed' indicates that this object
        has changed, notify all its observers, then
        call clearChanged(). Each observer is called directly'''
       
        if not self.changed: return
        # additions of observers:
        self.clear_changed()
        for observer in self.obs:
            #observer.update(self, arg)
            observer(arg)

    def delete_observers(self): self.obs = []
    def set_changed(self): self.changed = 1
    def clear_changed(self): self.changed = 0
    def has_changed(self): return self.changed
    def count_observers(self): return len(self.obs)

class Notifier(object):
    """Object that notfies observers when value changes.
    Either triggers on is/is not or on any change."""
    def __init__(self):
        self.value = None
        self.nots_is = {}
        self.nots_is_not = self.GeneralNotifier(self)
        self.nots_changed = self.GeneralNotifier(self)
        self.previousvalue = None
        self.notify_value_change = lambda _: True # no action, unless observers added
    
    def notify_value_change_is(self, value):
        """Nofifies observers if value is, otherwise notifies other observers."""
        if value in self.nots_is:
            self.nots_is[self.value].notify(self)
        else:
            self.nots_is_not.notify(self)
            
    def notify_value_change_changed(self, _):
        """Notifies obersers on any change."""
        self.nots_changed.notify(self)
        
    class GeneralNotifier(Observable):
        """Notifier which only triggers on change of outer value"""
        def __init__(self, outer):
            Observable.__init__(self)
            self.previousvalue = None
            self.outer = outer

        def notify(self, arg=None):
            """Notify if changed."""
            if not self.outer.value == self.outer.previousvalue:
                self.set_changed()
                Observable.notify_observers(self, arg)
                self.outer.previousvalue = self.outer.value

    def add_notifable_is(self, value, method):
        """Add notifable for value is."""
        self.nots_is.setdefault(value, self.GeneralNotifier(self)).add_observer(method)
        self.notify_value_change = self.notify_value_change_is
    def delete_notifable_is(self, value, method):
        """Remove notifiable."""
        self.nots_is.get(value, self.GeneralNotifierCompare(self)).delete_observer(method)
    def add_notifable_is_not(self, method):
        """Add notifable for value is not."""
        self.notify_value_change = self.notify_value_change_is
        self.nots_is_not.add_observer(method)
    def delete_notifable_is_not(self, method):
        """Remove notifiable."""
        self.nots_is_not.delete_dbserver(method)
    def add_notifable_changed(self, method):
        """Add notifable for value changes."""
        self.nots_changed.add_observer(method)
        self.notify_value_change = self.notify_value_change_changed
    def delete_notifable_changed(self, method):
        """Remove notifiable."""
        self.nots_changed.delete_observer(method)



"""Thermostat statemachine to represent heat controller in Heatmiser ThermoStats

Ian Horsley 2018
"""

from transitions import Machine

from hm_constants import MAX_AGE_MEDIUM
from schedule_functions import SCH_ENT_TEMP

class Thermostat(object):
    """Thermostat statemachine"""
    states = [{'name': 'off', 'on_enter': 'thres_off'},
            {'name': 'offfrost', 'on_enter': 'thres_frost'},
            {'name': 'frost', 'on_enter': 'thres_frost'},
            {'name': 'setpoint', 'on_enter': 'thres_setpoint'}
            ]
    
    TEMP_STATE_OFF = 0    #thermostat display is off and frost protection disabled
    TEMP_STATE_OFF_FROST = 1 #thermostat display is off and frost protection enabled
    TEMP_STATE_FROST = 2 #frost protection enabled indefinitely
    TEMP_STATE_HOLIDAY = 3 #holiday mode, frost protection for a period
    TEMP_STATE_HELD = 4 #temperature held for a number of hours
    TEMP_STATE_OVERRIDDEN = 5 #temperature overridden until next program time
    TEMP_STATE_PROGRAM = 6 #following program
    
    def thres_off(self, _=None):
        """Entry to off state, set threshold to None and set text."""
        print("STATE off")
        self.threshold = None
        self.text_function = lambda _: "controller off without frost protection"
        
    def thres_setpoint(self, _=None):
        """Entry to setpoint state, set threshold to setpoint and set text override, hold or program."""
        print("STATE to setpoint ", self.fieldscont.setroomtemp.value)
        self.threshold = self.fieldscont.setroomtemp.value
        
        if self.fieldscont.tempholdmins.value != 0:
            self.text_function = lambda infields: "temp held for %i mins at %i"%(infields.tempholdmins.value, infields.setroomtemp.value)
        else:
            self.text_function = self._text_function_over_prog

    def thres_frost(self, _=None):
        """Entry to frost state, set threshold to frost and set text off, frost or holiday."""
        print("STATE to", self.state)
        
        self.threshold = self.fieldscont.frosttemp.value
        
        if self.fieldscont.onoff.is_value('OFF'):
            self.text_function = lambda _: "controller off"
        elif self.fieldscont.holidayhours.value != 0:
            self.text_function = lambda infields: "controller on holiday for %s hours" % (infields.holidayhours.value)
        elif self.fieldscont.runmode.is_value('FROST'):
            self.text_function = lambda _: "controller in frost mode"
        else:
            self.text_function = None
    
    @staticmethod
    def _text_function_over_prog(infields):
        infields.read_field('currenttime', MAX_AGE_MEDIUM)
            
        locatimenow = infields.currenttime.localtimearray()
        scheduletarget = infields.heat_schedule.get_current_schedule_item(locatimenow)

        if infields.setroomtemp.value != scheduletarget[SCH_ENT_TEMP]:
            basetext = "temp overridden"
        else:
            basetext = "temp set"
        return basetext + "temp overridden to %0.1f until %02d:%02d" % (infields.setroomtemp.value, infields.nexttarget()[1], infields.nexttarget()[2])
    
    def get_state_text(self):
        """Return text desription of current state."""
        return self.text_function(self.fieldscont)
    
    def cond_frost(self, _=None):
        """Check is no frost triggers set."""
        return self.fieldscont.holidayhours.is_value('OFF') and self.fieldscont.runmode.is_value('HEAT')

    def cond_on(self, _=None):
        """Check switched on."""
        return self.fieldscont.onoff.is_value('ON')
        
    def cond_frostprotdisable(self, _=None):
        """Check frost protection is disabled."""
        return self.fieldscont.frostprotdisable.is_value('ON')
       
    def __init__(self, name, fieldcontainer):
    
        # intialise variables
        self.name = name #name of thermostat
        self.threshold = None #to store current thershold value
        self.text_function = lambda _: "unknown state"
        
        # field pointers
        self.fieldscont = fieldcontainer
        
        self.machine = Machine(model=self, states=Thermostat.states, initial='off')
        
        self.machine.add_transition('switch_off', ['frost', 'setpoint'], 'off', conditions='cond_frostprotdisable')
        self.machine.add_transition('switch_off', ['frost', 'setpoint'], 'offfrost', unless='cond_frostprotdisable')
        self.machine.add_transition('switch_off', 'offfrost', 'off', conditions='cond_frostprotdisable')
        self.machine.add_transition('switch_off', 'off', 'offfrost', unless='cond_frostprotdisable')

        self.machine.add_transition('switch_swap', '*', 'setpoint', conditions=['cond_frost', 'cond_on'])
        self.machine.add_transition('switch_swap', '*', 'frost', conditions='cond_on', unless='cond_frost')
        


#!/usr/bin/env python
"""Gets data from Heatmiser stats repeatedly and displays
Ian Horsley 2018
"""

import time
import logging

from heatmisercontroller.logging_setup import initialize_logger_full
from heatmisercontroller.network import HeatmiserNetwork
from heatmisercontroller.exceptions import HeatmiserResponseError

#start logging
initialize_logger_full('logs', logging.WARN)

HMN = HeatmiserNetwork()

# CYCLE THROUGH ALL CONTROLLERS
for current_controller in HMN.controllers:
    print("\r\nGetting all data control %2d in %s *****************************" %(current_controller.set_address, current_controller.set_long_name))

    try:
        current_controller.read_all()
        disptext = "C%d Air Temp is %.1f from type %.f and Target set to %d    Boiler Demand %d" % (current_controller.address, current_controller.read_air_temp(), current_controller.read_air_sensor_type(), current_controller.setroomtemp, current_controller.heatingdemand)
    except HeatmiserResponseError as err:
        print("C%d in %s Failed to Read due to %s" %(current_controller.address, current_controller.name.ljust(4), str(err)))
    else:
        if current_controller.is_hot_water:
            print("%s Hot Water Demand %d" %(disptext, current_controller.hotwaterdemand))
            current_controller.display_water_schedule()
        else:
            print(disptext)
        current_controller.display_heating_schedule()
        

time.sleep(5) # sleep before next cycle

#list fields used in following loop
SCHEDULEFIELDS = ['mon_heat', 'tues_heat', 'wed_heat', 'thurs_heat', 'fri_heat', 'wday_heat', 'wend_heat']
STATEFIELDS = ['onoff', 'frostprot', 'holidayhours', 'runmode', 'tempholdmins', 'setroomtemp']
OTHERFIELDS = ['sensorsavaliable', 'airtemp', 'remoteairtemp', 'heatingdemand', 'hotwaterdemand']
FIELDNAMES = STATEFIELDS + OTHERFIELDS

# CYCLE THROUGH ALL CONTROLLERS repeatedly, constantly updating parameters
while True:
    print()
    for current_controller in HMN.controllers:
        try:
            #read all fields at the same time allows for most efficient number of reads
            current_controller.read_fields(SCHEDULEFIELDS)
            current_controller.read_fields(FIELDNAMES, 0) #force get on these fields
            targettext = current_controller.print_target()
            disptext = "C%d in %s Air Temp is %.1f from type %.f, %s, Heat %d" %(current_controller.address, current_controller.name.ljust(4), current_controller.read_air_temp(), current_controller.read_air_sensor_type(), targettext, current_controller.heatingdemand)
        except HeatmiserResponseError as err:
            print("C%d in %s Failed to Read due to %s" %(current_controller.address, current_controller.name.ljust(4), str(err)))
        else:
            if current_controller.is_hot_water:
                print("%s Water %d" % (disptext, current_controller.hotwaterdemand))
            else:
                print(disptext)

    time.sleep(5) # sleep before next cycle



#!/usr/bin/env python
"""Gets data from Heatmiser stats repeatedly and displays
Ian Horsley 2018
"""
import os
import logging

from heatmisercontroller.logging_setup import initialize_logger_full
from heatmisercontroller.network import HeatmiserNetwork
from heatmisercontroller.exceptions import HeatmiserResponseError

#start logging
initialize_logger_full('logs', logging.INFO)

MODULE_PATH = os.path.abspath(os.path.dirname(__file__))
CONFIGFILE = os.path.join(MODULE_PATH, "nocontrollers.conf")

HMN = HeatmiserNetwork(CONFIGFILE)
HMN.find_devices(10)

# CYCLE THROUGH ALL CONTROLLERS
for current_controller in HMN.controllers:
    print("\r\nGetting all data control %2d in %s *****************************" %(current_controller.address, current_controller.long_name))

    try:
        current_controller.read_all()
    except HeatmiserResponseError as err:
        print("C%d in %s Failed to Read due to %s" %(current_controller.address, current_controller.name.ljust(4), str(err)))
    else:
        disptext = "C%d Air Temp is %.1f from type %.f and Target set to %d    Boiler Demand %d" % (current_controller.address, current_controller.read_air_temp(), current_controller.read_air_sensor_type(), current_controller.setroomtemp, current_controller.heatingdemand)
        if current_controller.is_hot_water:
            print("%s Hot Water Demand %d" %(disptext, current_controller.hotwaterdemand))
            current_controller.display_water_schedule()
        else:
            print(disptext)
        current_controller.display_heating_schedule()



"""field definitions for Heatmiser protocol"""
import logging
import time

from hm_constants import CURRENT_TIME_DAY, CURRENT_TIME_HOUR, CURRENT_TIME_MIN, CURRENT_TIME_SEC, TIME_ERR_LIMIT
from hm_constants import BYTEMASK
from .exceptions import HeatmiserResponseError, HeatmiserControllerTimeError
from .observer import Notifier

VALUES_ON_OFF = {'ON': 1, 'OFF': 0} #assusme that default comes first, need to swtich to ordered dictionary to make it possible to get default value
VALUES_OFF_ON = {'OFF': 0, 'ON': 1}
VALUES_OFF = {'OFF': 0}

class HeatmiserFieldUnknown(Notifier):
    """Class for variable length unknown read only field"""
    writeable = False
    divisor = 1
    
    def __init__(self, name, address, max_age, length):
        super(HeatmiserFieldUnknown, self).__init__()
        self.name = name
        self.address = address
        self.dcbaddress = address
        self.max_age = max_age
        self.fieldlength = length
        self._reset()
    
    def __eq__(self, other):
        return self.value == other
        
    def __int__(self):
        return self.value

    def __str__(self):
        return str(self.value)
        
    def __cmp__(self, value):
        return cmp(self.value, value)

    def _reset(self):
        """Reset data and values to unknown."""
        self.data = None
        self.value = None
        self.lastreadtime = None #used to record when the field was last read
    
    def last_dcb_byte_address(self):
        """returns the address of the last dcb byte"""
        return self.dcbaddress + self.fieldlength - 1
    
    def update_data(self, data, readtime):
        """update stored data and readtime. Don't compute value because don't know how to map"""
        self.data = data
        self.lastreadtime = readtime
        return data
    
    def update_value(self, value, writetime):
        """Don't update because don't know how to map to data."""
        raise NotImplementedError

    def is_writable(self):
        """Checks if field is writable"""
        if not self.writeable:
            raise ValueError("set_field: field isn't writeable")

    def check_values(self, values):
        """check a payload matches field spec"""
        raise NotImplementedError
        
    def check_data_valid(self):
        """check whether data has been set"""
        data_not_valid = self.lastreadtime == None
        if data_not_valid:
            logging.debug("Data item %s not available"%(self.name))
        return not data_not_valid
        
    def check_data_fresh(self, maxagein=None):
        """check whether data is fresh
        
        Check field data age is not more than maxage (in seconds)
        maxagein = None, use the default self.maxage
        maxagein = -1, only check if present
        maxagein >=0, use maxagein (0 is effectively always False)
        return False if old, True if recent"""
        if not self.check_data_valid():
            return False
        elif maxagein == -1: #only check present
            return True
        elif maxagein == None: #if none use field defaults
            maxage = self.max_age
        else:
            maxage = maxagein
        #now check time
        if time.time() - self.lastreadtime > maxage:
            logging.debug("Data item %s too old"%(self.name))
            return False
        return True
        
class HeatmiserField(HeatmiserFieldUnknown):
    """Base class for fields providing basic method calls"""
    #single value and hence single range
    writeable = True
    
    def __init__(self, name, address, validrange, max_age, readvalues=None):
		###valid range list can be [], [min, max], [list of valid values]
        super(HeatmiserField, self).__init__(name, address, max_age, self.fieldlength)
        self.validrange = validrange
        self.value = None
        self.expectedvalue = None
        self.writevalues = self.readvalues = readvalues
        #check isinstance(fieldrange[0], (int, long)) and isinstance(fieldrange[1], (int, long))
        if len(validrange) < 2:
            self.validrange = [0, self.maxdatavalue / self.divisor]
    
    def is_value(self, name):
        """Returns true if value matches read value."""
        return self.value == self.readvalues[name]
    
    def read_value_text(self):
        """returns value converting to label if known"""
        if self.readvalues is None:
            return self.value
        else:
            return self.readvalues.keys()[self.readvalues.values().index(self.value)]
            
    def write_value_from_text(self, value):
        """maps text to value, otherwise returns input"""
        if self.writevalues is None:
            return value
        else:
            return self.writevalues.get(value, value)
    
    def update_data(self, data, readtime):
        """update stored data and readtime if data valid. Compute and store value from data."""
        value = self._calculate_value(data)
        if not self.expectedvalue is None and value != self.expectedvalue:
            raise HeatmiserResponseError('Value %i is unexpected for %s, expected %i'%(value, self.name, self.expectedvalue))
        self._validate_range(value)
        self.data = data
        self.value = value
        self.lastreadtime = readtime
        self.notify_value_change(value)
        return value

    def update_value(self, value, writetime):
        """Update the field value once successfully written to network"""
        self._validate_range(value, ValueError)
        data = self.format_data_from_value(value)
        self.data = data
        self.value = value
        self.lastreadtime = writetime
        self.notify_value_change(value)
        return value
        
    def _validate_range(self, values, errortype=HeatmiserResponseError, range=None):
        """validate the value is within range or in list."""
        if range is None:
            range = self.validrange

        if len(range) == 2:
            if values < range[0] or values > range[1]:
                raise errortype("Value %.1f  outside expected range (%.1f, %.1f) for %s"%(values, range[0], range[1], self.name))
        elif len(range) > 2:
            if values not in range:
                raise errortype("Value %.1f  outside expected range %s for %s"%(values, ','.join(map(str, range)), self.name))
        else:
            raise errortype("Expected range not defined for %s"%(self.name))

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        raise NotImplementedError
        
    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        raise NotImplementedError
        
    def check_values(self, values):
        """check a single or double byte field value matches field spec"""
        if not isinstance(values, (int, long)):
            #one or two byte field, not single length values
            raise TypeError("set_field: invalid requested value")

        #checks the values matches the ranges if ranges are defined
        self._validate_range(values, ValueError)

class HeatmiserFieldSingle(HeatmiserField):
    """Class for writable 1 byte field"""
    maxdatavalue = 255
    fieldlength = 1

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        return data[0]/self.divisor
    
    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        return [value]

class HeatmiserFieldSingleReadOnly(HeatmiserFieldSingle):
    """Class for read only 1 byte field"""
    writeable = False

class HeatmiserFieldHotWaterVersion(HeatmiserFieldSingleReadOnly):
    """Class for version on hotwater models."""
    floorlimiting = None
        
    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        self.floorlimiting = data[0] >> 7
        return data[0] & 0x7f
        
class HeatmiserFieldHotWaterDemand(HeatmiserFieldSingle):
    """Class to impliment read and write differences for hotwater demand field."""
    def __init__(self, name, address, validrange, max_age):
        super(HeatmiserFieldHotWaterDemand, self).__init__(name, address, validrange, max_age, VALUES_ON_OFF)
        self.writevalues = {'PROG': 0, 'OVER_ON': 1, 'OVER_OFF': 2}

    def update_value(self, value, writetime):
        """Update the field value once successfully written to network if known. Otherwise reset"""
        #handle odd effect on WRITE_hotwaterdemand_PROG
        
        if value == self.writevalues['PROG']: #returned to program so outcome is unknown
            self._reset()
            return None
        elif value == self.writevalues['OVER_OFF']: #if overridden off store the off read value
            return super(HeatmiserFieldHotWaterDemand, self).update_value(self.readvalues['OFF'], writetime)
        else:
            return super(HeatmiserFieldHotWaterDemand, self).update_value(value, writetime)
        
class HeatmiserFieldDouble(HeatmiserField):
    """Class for writable 2 byte field"""
    maxdatavalue = 65535
    fieldlength = 2
        
    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        val_high = data[0]
        val_low = data[1]
        return 1.0*(val_high*256 + val_low)/self.divisor #force float, although always returns integer temps.

    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        pay_lo = (value & BYTEMASK)
        pay_hi = (value >> 8) & BYTEMASK
        return [pay_lo, pay_hi]
        
class HeatmiserFieldDoubleReadOnly(HeatmiserFieldDouble):
    """Class for read only 2 byte field"""
    writeable = False

class HeatmiserFieldDoubleReadOnlyTenths(HeatmiserFieldDoubleReadOnly):
    """Class for read only 2 byte field"""
    divisor = 10.0
        
class HeatmiserFieldMulti(HeatmiserField):
    """Base class for writable multi byte field"""
    maxdatavalue = None
        
    def _validate_range(self, values, errortype=HeatmiserResponseError):
        """validate the value is within range or in list. cyles through list of ranges"""
        for i, item in enumerate(values):
            range = self.validrange[i % len(self.validrange)]
            super(HeatmiserFieldMulti, self)._validate_range(item, errortype, range)

    def _calculate_value(self, data):
        """Calculate value from payload bytes"""
        return data

    def format_data_from_value(self, value):
        """Convert field to byte form for writting to device"""
        return list(value) #force a copy
    
    def check_values(self, values):
        """check a values matches field spec"""
        if len(values) != self.fieldlength:
            #greater than two byte field, values length must match field length
            raise ValueError("set_field: invalid payload length")

        #checks the values matches the ranges if ranges are defined
        self._validate_range(values, ValueError)

class HeatmiserFieldTime(HeatmiserFieldMulti):
    """Class for time field"""
    fieldlength = 4
    
    def __init__(self, name, address, validrange, max_age):
        self.timeerr = None
        super(HeatmiserFieldTime, self).__init__(name, address, validrange, max_age)
        
    def comparecontrollertime(self):
        """Compare device and local time difference against threshold"""
        # Now do same sanity checking
        # Check the time is within range
        # currentday is numbered 1-7 for M-S
        # localday (python) is numbered 0-6 for Sun-Sat
        
        if not self.check_data_valid():
            raise HeatmiserResponseError("Time not read before check")

        localtimearray = self.localtimearray(self.lastreadtime) #time that time field was read
        localweeksecs = self._weeksecs(localtimearray)
        remoteweeksecs = self._weeksecs(self.value)
        directdifference = abs(localweeksecs - remoteweeksecs)
        wrappeddifference = abs(self.DAYSECS * 7 - directdifference) #compute the difference on rollover
        self.timeerr = min(directdifference, wrappeddifference)
        logging.debug("Local time %i, remote time %i, error %i"%(localweeksecs, remoteweeksecs, self.timeerr))

        if self.timeerr > self.DAYSECS:
            raise HeatmiserControllerTimeError("C%2d Incorrect day : local is %s, sensor is %s" % (self.address, localtimearray[CURRENT_TIME_DAY], self.value[CURRENT_TIME_DAY]))

        if self.timeerr > TIME_ERR_LIMIT:
            raise HeatmiserControllerTimeError("C%2d Time Error %d greater than %d: local is %s, sensor is %s" % (self.address, self.timeerr, TIME_ERR_LIMIT, localweeksecs, remoteweeksecs))

    @staticmethod
    def localtimearray(timenow=time.time()):
        """creates an array in heatmiser format for local time. Day 1-7, 1=Monday"""
        #input time.time() (not local)
        localtimenow = time.localtime(timenow)
        nowday = localtimenow.tm_wday + 1 #python tm_wday, range [0, 6], Monday is 0
        nowsecs = min(localtimenow.tm_sec, 59) #python tm_sec range[0, 61]
        
        return [nowday, localtimenow.tm_hour, localtimenow.tm_min, nowsecs]
    
    DAYSECS = 86400
    HOURSECS = 3600
    MINSECS = 60

    def _weeksecs(self, localtimearray):
        """calculates the time from the start of the week in seconds from a heatmiser time array"""
        return (localtimearray[CURRENT_TIME_DAY] - 1) * self.DAYSECS + localtimearray[CURRENT_TIME_HOUR] * self.HOURSECS + localtimearray[CURRENT_TIME_MIN] * self.MINSECS + localtimearray[CURRENT_TIME_SEC]

class HeatmiserFieldHeat(HeatmiserFieldMulti):
    """Class for heating schedule field"""
    fieldlength = 12

class HeatmiserFieldWater(HeatmiserFieldMulti):
    """Class for hotwater schedule field"""
    fieldlength = 16



"""Constants and """

# Protocols
HMV2_ID = 2
HMV3_ID = 3
DEFAULT_PROTOCOL = HMV3_ID

BYTEMASK = 0xff

DONT_CARE_LENGTH = 1
TIME_ERR_LIMIT = 10

#
# HM Version 3 Magic Numbers
#

# Master must be in range [0x81, 0xa0] = [129, 160]
MASTER_ADDR_MIN = 0x81
MASTER_ADDR_MAX = 0xa0
SLAVE_ADDR_MIN = 1
SLAVE_ADDR_MAX = 32

MAX_FRAME_RESP_LENGTH = 159 # NB max return is 75 in 5/2 mode or 159 in 7day mode
MIN_FRAME_RESP_LENGTH = 7
FRAME_WRITE_RESP_LENGTH = 7
MIN_FRAME_READ_RESP_LENGTH = 11
MIN_FRAME_SEND_LENGTH = 10
MAX_PAYLOAD_SEND_LENGTH = 100
CRC_LENGTH = 2

# Define magic numbers used in messages
FUNC_READ = 0
FUNC_WRITE = 1

BROADCAST_ADDR = 0xff
RW_LENGTH_ALL = 0xffff
DCB_START = 0x00

# Define frame send contents
FS_DEST_ADDR = 0
FS_LEN = 1
FS_SOURCE_ADDR = 2
FS_FUNC_CODE = 3

# Define frame response contents
FR_DEST_ADDR = 0
FR_LEN_LOW = 1
FR_LEN_HIGH = 2
FR_SOURCE_ADDR = 3
FR_FUNC_CODE = 4
FR_START_LOW = 5 # this field and following on present if function code read
FR_START_HIGH = 6
FR_CONT_LEN_LOW = 7
FR_CONT_LEN_HIGH = 8
FR_CONTENTS = 9

MAX_AGE_LONG = 86400
MAX_AGE_MEDIUM = 3600
MAX_AGE_SHORT = 65
MAX_AGE_USHORT = 11

FIELD_NAME_LENGTH = 13
MAX_UNIQUE_ADDRESS = 298

CURRENT_TIME_DAY = 0
CURRENT_TIME_HOUR = 1
CURRENT_TIME_MIN = 2
CURRENT_TIME_SEC = 3

#these maps are used to translate the unique address which is the same for all controllers to the DCB address for a specific controller
PROG_MODE_WEEK = 'week'
PROG_MODE_DAY = 'day'
DEFAULT_PROG_MODE = PROG_MODE_DAY #allows broadcast to program both modes.

#Fields for each stat type and mode,  list of ranges.
#PRT-E fields
FIELDRANGES = {}
FIELDRANGES['prt_e_model'] = {}
FIELDRANGES['prt_e_model'][PROG_MODE_WEEK] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'heatingdemand'], ['currenttime', 'wend_heat']]
FIELDRANGES['prt_e_model'][PROG_MODE_DAY] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'heatingdemand'], ['currenttime', 'wend_heat'], ['mon_heat', 'sun_heat']]
#PRT-HW fields
FIELDRANGES['prt_hw_model'] = {}
FIELDRANGES['prt_hw_model'][PROG_MODE_WEEK] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'wend_water']]
FIELDRANGES['prt_hw_model'][PROG_MODE_DAY] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'sun_water']]
#Full map
FIELDRANGES[False] = {PROG_MODE_WEEK: [['DCBlen', 'sun_water']], PROG_MODE_DAY: [['DCBlen', 'sun_water']]}



#!/usr/bin/env python
"""Example of setting configurations on Heatmiser stats
Ian Horsley 2018
"""

import logging

from heatmisercontroller.logging_setup import initialize_logger_full
from heatmisercontroller.network import HeatmiserNetwork

initialize_logger_full('logs', logging.INFO)

HMN = HeatmiserNetwork()

### New observations
# Setting holiday replaces overide and on exit returns to prog (not override)

#HMN.Kit.holdTemp(30,21) #mins, temp
#HMN.All.release_hold_temp()

#HMN.release_temp("Cons")
#HMN.All.set_field('holidayhours', 115)
#HMN.All.release_holiday()

#HMN.Kit.set_field('hotwaterdemand', 'PROG')

#HMN.All.set_field('onoff', 'ON')
#HMN.All.set_field('runmode', 'FROST')
#HMN.hmUpdateTime(2)

HMN.B1.set_field('runmode', 'HEAT')
HMN.B2.set_field('runmode', 'HEAT')
HMN.Kit.set_field('runmode', 'HEAT')
HMN.Cons.set_field('runmode', 'FROST')
HMN.Sit.set_field('runmode', 'HEAT')

HMN.All.set_field('frosttemp', 10)
HMN.Cons.set_field('frosttemp', 7)

print HMN.Cons.read_field('frostprotdisable')
HMN.Cons.set_field('onoff', 'OFF')

#print HMN.All.read_field('holidayhours', 0)
#exit()


#HMN.B2.set_temp(24)
#HMN.Kit.set_temp(24)
#HMN.B1.set_temp(24)
#HMN.Cons.set_temp(24)
#HMN.All.set_temp(24)

#HMN.Kit.release_temp()
#HMN.B1.release_temp()
#HMN.B2.release_temp()
#HMN.Cons.release_temp()

DOWNINACTIVE = 20
DOWNACTIVE = 19
FROST = 12
SLEEP = 16
DAYFROST = [7, 0, FROST]

WKDAY_ZEARLY = [5, 00, DOWNACTIVE, 8, 30, FROST, 15, 00, DOWNINACTIVE, 21, 00, SLEEP]
WKDAY_ZLATE = [7, 00, DOWNINACTIVE, 12, 00, FROST, 17, 00, DOWNINACTIVE, 21, 30, SLEEP]
WKDAY_ZTRAINING = [6, 30, DOWNACTIVE, 8, 30, FROST, 16, 30, DOWNINACTIVE, 21, 00, SLEEP]
WKEND_ZEARLY = [5, 00, DOWNINACTIVE, 21, 30, SLEEP]
WKEND_ZLATE = [7, 00, DOWNINACTIVE, 21, 30, SLEEP]
WKEND_ZOFF = [7, 00, DOWNINACTIVE, 21, 30, SLEEP]
WKDAY_ZAWAY = [7, 00, DOWNACTIVE, 8, 30, FROST, 17, 30, DOWNINACTIVE, 21, 00, SLEEP]
WKDAY_ZSTARTNIGHT = WKEND_ZOFF

HMN.Kit.set_heating_schedule('mon_heat', WKDAY_ZAWAY)
HMN.Kit.set_heating_schedule('tues_heat', WKDAY_ZAWAY)
HMN.Kit.set_heating_schedule('wed_heat', WKDAY_ZAWAY)
HMN.Kit.set_heating_schedule('thurs_heat', WKDAY_ZAWAY)
HMN.Kit.set_heating_schedule('fri_heat', WKDAY_ZAWAY)
HMN.Kit.set_heating_schedule('sat_heat', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('sun_heat', WKEND_ZOFF)

HMN.Sit.set_heating_schedule('mon_heat', WKDAY_ZAWAY)
HMN.Sit.set_heating_schedule('tues_heat', WKDAY_ZAWAY)
HMN.Sit.set_heating_schedule('wed_heat', WKDAY_ZAWAY)
HMN.Sit.set_heating_schedule('thurs_heat', WKDAY_ZAWAY)
HMN.Sit.set_heating_schedule('fri_heat', WKDAY_ZAWAY)
HMN.Sit.set_heating_schedule('sat_heat', WKEND_ZOFF)
HMN.Sit.set_heating_schedule('sun_heat', WKEND_ZOFF)

EVENINGWATER = [17, 30, 18, 30]
NOWATER = [24, 00, 21, 30]
HMN.Kit.set_water_schedule('all', NOWATER)

UPSLEEP = 12
UPAWAKE = 15
UPWKDAY_ZEARLY = [5, 0, UPAWAKE, 7, 00, FROST, 15, 00, UPAWAKE, 19, 30, UPSLEEP]
UPWKDAY_ZLATE = [7, 0, UPAWAKE, 8, 00, FROST, 17, 00, UPAWAKE, 20, 30, UPSLEEP]
UPWKDAY = [7, 0, UPAWAKE, 8, 30, FROST, 21, 00, UPAWAKE, 21, 30, UPSLEEP]
UPWKEND_ZEARLY = [5, 0, UPAWAKE, 8, 00, FROST, 15, 00, UPAWAKE, 19, 30, UPSLEEP]
UPWKEND_ZLATE = [7, 0, UPAWAKE, 9, 30, FROST, 17, 00, UPAWAKE, 20, 30, UPSLEEP]
UPWKEND = [7, 0, UPAWAKE, 9, 30, FROST, 17, 00, UPAWAKE, 19, 00, UPSLEEP]
DAYFROST = [7, 0, FROST]

HMN.B1.set_heating_schedule('mon_heat', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('tues_heat', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('wed_heat', UPWKDAY_ZLATE)
HMN.B1.set_heating_schedule('thurs_heat', UPWKDAY_ZEARLY)
HMN.B1.set_heating_schedule('fri_heat', UPWKDAY_ZEARLY)
HMN.B1.set_heating_schedule('sat_heat', DAYFROST)
HMN.B1.set_heating_schedule('sun_heat', DAYFROST)

HMN.B2.set_heating_schedule('mon_heat', UPWKDAY_ZLATE)
HMN.B2.set_heating_schedule('tues_heat', UPWKDAY_ZLATE)
HMN.B2.set_heating_schedule('wed_heat', UPWKDAY_ZLATE)
HMN.B2.set_heating_schedule('thurs_heat', UPWKDAY_ZEARLY)
HMN.B2.set_heating_schedule('fri_heat', UPWKDAY_ZEARLY)
HMN.B2.set_heating_schedule('sat_heat', DAYFROST)
HMN.B2.set_heating_schedule('sun_heat', DAYFROST)

HMN.get_controller_by_name('Cons').set_field('wday_heat', [9, 0, 12, 21, 30, 10, 24, 0, 5, 24, 0, 5])
HMN.get_controller_by_name('Cons').set_field('wend_heat', [9, 0, 12, 21, 30, 10, 24, 0, 5, 24, 0, 5])

HMN.Sit.set_heating_schedule('wday_heat', DAYFROST)
HMN.Sit.set_heating_schedule('wend_heat', DAYFROST)

#HMN.hmSetFields("Kit", 'wday_water', [7, 0, 8, 0, 16, 0, 17, 0, 24, 0, 24, 0, 24, 0, 24, 0])
#HMN.hmSetFields("Kit", 'wday_water', [24, 0, 24, 0, 24, 0, 24, 0, 24, 0, 24, 0, 24, 0, 24, 0])
#HMN.hmSetFields("Kit", 'wend_water', [8, 0, 9, 0, 18, 0, 19, 0, 24, 0, 24, 0, 24, 0, 24, 0])

#HMN.hmSetField("Kit", 'frosttemp', 9)
#HMN.hmSetField("B1", 'frosttemp', 9)
#HMN.hmSetField("B2", 'frosttemp', 9)
#HMN.hmSetField("Cons", 'frosttemp', 9)





"""Heatmiser Device Classes

PRT-E Thermostat classes on the Heatmiser network

Ian Horsley 2018
"""
import logging
import time

from genericdevice import HeatmiserDevice, DEVICETYPES
from fields import HeatmiserFieldSingle, HeatmiserFieldSingleReadOnly, HeatmiserFieldDouble, HeatmiserFieldDoubleReadOnly, HeatmiserFieldTime, HeatmiserFieldHeat, HeatmiserFieldDoubleReadOnlyTenths
from fields import VALUES_ON_OFF, VALUES_OFF_ON, VALUES_OFF
from hm_constants import MAX_AGE_LONG, MAX_AGE_MEDIUM, MAX_AGE_SHORT, MAX_AGE_USHORT
from .exceptions import HeatmiserControllerTimeError, HeatmiserControllerSensorError
from schedule_functions import SchedulerDayHeat, SchedulerWeekHeat
from thermostatstate import Thermostat

class ThermoStatWeek(HeatmiserDevice):
    """Device class for thermostats operating weekly programmode
    Heatmiser prt_e_model."""
    is_hot_water = False #returns True if stat is a model with hotwater control, False otherwise
    
    def __init__(self, adaptor, devicesettings, generalsettings={}):
        self.heat_schedule = None #placeholder for heating schedule object
        self.thermostat = None #placeholder for thermostat object
        super(ThermoStatWeek, self).__init__(adaptor, devicesettings, generalsettings)
        #thermostat specific
    
    def _buildfields(self):
        """add to list of fields"""
        super(ThermoStatWeek, self)._buildfields()
        # list of fields can be sorted by key
        # dcb addresses could be computed from the completed field list and added to field.
        # all should have the first 4 fields, so put these in generic
        self.fields.extend([
            HeatmiserFieldSingleReadOnly('tempformat', 5, [0, 1], MAX_AGE_LONG),  # 00 C,  01 F
            HeatmiserFieldSingleReadOnly('switchdiff', 6, [1, 3], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('frostprotdisable', 7, [0, 1], MAX_AGE_LONG, VALUES_OFF_ON),  #0=enable frost prot when display off,  (opposite in protocol manual,  but tested and user guide is correct)  (default should be enabled)
            HeatmiserFieldDoubleReadOnly('caloffset', 8, [], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('outputdelay', 10, [0, 15], MAX_AGE_LONG),  # minutes (to prevent rapid switching)
            HeatmiserFieldSingleReadOnly('updwnkeylimit', 12, [0, 10], MAX_AGE_LONG),   #limits use of up and down keys
            HeatmiserFieldSingleReadOnly('sensorsavaliable', 13, [0, 4], MAX_AGE_LONG, {'INT_ONLY': 0, 'EXT_ONLY': 1, 'FLOOR_ONLY': 2, 'INT_FLOOR': 3, 'EXT_FLOOR': 4}),  #00 built in only,  01 remote air only,  02 floor only,  03 built in + floor,  04 remote + floor
            HeatmiserFieldSingleReadOnly('optimstart', 14, [0, 3], MAX_AGE_LONG),  # 0 to 3 hours,  default 0
            HeatmiserFieldSingleReadOnly('rateofchange', 15, [], MAX_AGE_LONG),  #number of minutes per degree to raise the temperature,  default 20. Applies to the Wake and Return comfort levels (1st and 3rd)
            HeatmiserFieldSingleReadOnly('programmode', 16, [0, 1], MAX_AGE_LONG, {'day':1, 'week':0}),  #0=5/2,  1= 7day
            HeatmiserFieldSingle('frosttemp', 17, [7, 17], MAX_AGE_LONG),  #default is 12,  frost protection temperature
            HeatmiserFieldSingle('setroomtemp', 18, [5, 35], MAX_AGE_USHORT),
            HeatmiserFieldSingle('floormaxlimit', 19, [20, 45], MAX_AGE_LONG),
            HeatmiserFieldSingleReadOnly('floormaxlimitenable', 20, [0, 1], MAX_AGE_LONG),  #1 is enable, 0 is disable
            HeatmiserFieldSingle('onoff', 21, [0, 1], MAX_AGE_SHORT, VALUES_ON_OFF),  #1 is on, 0 is off
            HeatmiserFieldSingle('keylock', 22, [0, 1], MAX_AGE_SHORT, VALUES_ON_OFF),  #1 is on, 0 is off
            HeatmiserFieldSingle('runmode', 23, [0, 1], MAX_AGE_SHORT, {'HEAT': 0, 'FROST': 1}),   #0 = heating mode,  1 = frost protection mode
            HeatmiserFieldDouble('holidayhours', 24, [0, 720], MAX_AGE_SHORT, VALUES_OFF),  #range guessed and tested,  setting to 0 cancels hold and puts back to program
            #HeatmiserFieldUnknown('unknown', 26, 1, [], MAX_AGE_LONG, 6),  # gap from 26 to 31
            HeatmiserFieldDouble('tempholdmins', 32, [0, 5760], MAX_AGE_SHORT, VALUES_OFF),  #range guessed and tested,  setting to 0 cancels hold and puts setroomtemp back to program
            HeatmiserFieldDoubleReadOnlyTenths('remoteairtemp', 34, [], MAX_AGE_USHORT),  #ffff if no sensor
            HeatmiserFieldDoubleReadOnlyTenths('floortemp', 36, [], MAX_AGE_USHORT),  #ffff if no sensor
            HeatmiserFieldDoubleReadOnlyTenths('airtemp', 38, [], MAX_AGE_USHORT),  #ffff if no sensor
            HeatmiserFieldSingleReadOnly('errorcode', 40, [0, 224, 225, 226], MAX_AGE_SHORT),  # 0 is no error # errors,  0 built in,  1,  floor,  2 remote
            HeatmiserFieldSingleReadOnly('heatingdemand', 41, [0, 1], MAX_AGE_USHORT),  #0 none,  1 heating currently
            HeatmiserFieldTime('currenttime', 43, [[1, 7], [0, 23], [0, 59], [0, 59]], MAX_AGE_USHORT),  #day (Mon - Sun),  hour,  min,  sec.
            #5/2 progamming #if hour = 24 entry not used
            HeatmiserFieldHeat('wday_heat', 47, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),  #hour,  min,  temp  (should minutes be only 0 and 30?)
            HeatmiserFieldHeat('wend_heat', 59, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM)
            ])

        self.heat_schedule = SchedulerWeekHeat()
        self.thermostat = Thermostat('Heating', self)
    
    def _connect_observers(self):
        """connect obersers to fields"""
        super(ThermoStatWeek, self)._connect_observers()
        self.wday_heat.add_notifable_changed(self.heat_schedule.set_raw_field)
        self.wend_heat.add_notifable_changed(self.heat_schedule.set_raw_field)
        
        #on and off
        self.frostprotdisable.add_notifable_is(self.frostprotdisable.readvalues['ON'], self.thermostat.switch_off)
        self.frostprotdisable.add_notifable_is(self.frostprotdisable.readvalues['OFF'], self.thermostat.switch_off)
        self.onoff.add_notifable_is(self.onoff.readvalues['OFF'], self.thermostat.switch_off)
        self.onoff.add_notifable_is(self.onoff.readvalues['ON'], self.thermostat.switch_swap)
        #change within on
        self.setroomtemp.add_notifable_changed(self.thermostat.switch_swap)
        ###don't need the following if not modelling override holds or program events internally
        ###program change event
        self.tempholdmins.add_notifable_is(self.tempholdmins.readvalues['OFF'], self.thermostat.switch_swap) #check triggered by value force change and data change
        self.tempholdmins.add_notifable_is_not(self.thermostat.switch_swap)
        #between frost and setpoint
        self.runmode.add_notifable_is(self.runmode.readvalues['HEAT'], self.thermostat.switch_swap)
        self.runmode.add_notifable_is(self.runmode.readvalues['FROST'], self.thermostat.switch_swap)
        self.holidayhours.add_notifable_is(self.holidayhours.readvalues['OFF'], self.thermostat.switch_swap)
        self.holidayhours.add_notifable_is_not(self.thermostat.switch_swap)

    def _set_expected_field_values(self):
        """set the expected values for fields that should be fixed"""
        super(ThermoStatWeek, self)._set_expected_field_values()
        self.programmode.expectedvalue = self.programmode.readvalues[self.set_expected_prog_mode]
    
    def _procfield(self, data, fieldinfo):
        """Process data for a single field storing in relevant."""
        super(ThermoStatWeek, self)._procfield(data, fieldinfo)
        
        if fieldinfo.name == 'version' and hasattr(fieldinfo, 'floorlimiting'):
            self.data['floorlimiting'] = self.floorlimiting
            
        if fieldinfo.name == 'currenttime':
            self._checkcontrollertime()
    
    def _checkcontrollertime(self):
        """run check of device time against local read time, and try to fix if _autocorrectime"""
        try:
            self.currenttime.comparecontrollertime()
        except HeatmiserControllerTimeError:
            if self.set_autocorrectime is True:
                ### Add warning that attempting to fix.
                self.set_time()
            else:
                raise
    
    def get_variables(self):
        """Gets setroomtemp to hotwaterdemand fields from device"""
        self.get_field_range('setroomtemp', 'hotwaterdemand')
        
    def get_temps_and_demand(self):
        """Gets remoteairtemp to hotwaterdemand fields from device"""
        self.get_field_range('remoteairtemp', 'hotwaterdemand')
    
    ## External functions for printing data
    def display_heating_schedule(self):
        """Prints heating schedule to stdout"""
        self.heat_schedule.display()

    def nexttarget(self):
        """get next heat target"""
        return self.heat_schedule.get_next_schedule_item(self.currenttime.localtimearray())

    def print_target(self):
        """Returns text describing current heating state"""
        self.read_temp_state()
        return self.thermostat.get_state_text()
            
    ## External functions for reading data
    
    def read_temp_state(self):
        """Returns the current temperature control state from off to following program"""
        self.read_fields(['mon_heat', 'tues_heat', 'wed_heat', 'thurs_heat', 'fri_heat', 'wday_heat', 'wend_heat'], -1)
        self.read_fields(['onoff', 'frostprotdisable', 'holidayhours', 'runmode', 'tempholdmins', 'setroomtemp'])

        return self.thermostat.state
        
    def read_air_sensor_type(self):
        """Reports airsensor type"""
        #1 local, 2 remote
        self.read_field('sensorsavaliable')

        if self.sensorsavaliable.is_value('INT_ONLY') or self.sensorsavaliable.is_value('INT_FLOOR'):
            return 1
        elif self.sensorsavaliable.is_value('EXT_ONLY') or self.sensorsavaliable.is_value('EXT_FLOOR'):
            return 2
        raise ValueError("sensorsavaliable field invalid")
            
    def read_air_temp(self):
        """Read the air temperature getting data from device if too old"""
        if self.read_air_sensor_type() == 1:
            self.read_fields(['airtemp', 'errorcode'], self.set_max_age_temp)
            if self.errorcode == 224:
                raise HeatmiserControllerSensorError("airtemp sensor error")
            return self.airtemp.value
        else:
            self.read_fields(['remoteairtemp', 'errorcode'], self.set_max_age_temp)
            if self.errorcode == 226:
                raise HeatmiserControllerSensorError("remote airtemp sensor error")
            return self.remoteairtemp.value
        
    def read_time(self, maxage=0):
        """Readtime, getting from device if required"""
        return self.read_field('currenttime', maxage)
        
    ## External functions for setting data

    def set_heating_schedule(self, day, schedule):
        """Set heating schedule for a single day"""
        padschedule = self.heat_schedule.pad_schedule(schedule)
        self.set_field(day, padschedule)

    def set_time(self):
        """set time on device to match current localtime on server"""
        timenow = time.time() + 0.5 #allow a little time for any delay in setting
        return self.set_field('currenttime', self.currenttime.localtimearray(timenow))

    #overriding

    def set_temp(self, temp):
        """sets the temperature demand overriding the program."""
        #Believe it returns at next prog change.
        if self.read_field('tempholdmins') == 0: #check hold temp not applied
            return self.set_field('setroomtemp', temp)
        else:
            logging.warn("%i address, temp hold applied so won't set temp"%(self.set_address))

    def release_temp(self):
        """release setTemp back to the program, but only if temp isn't held for a time (holdTemp)."""
        if self.read_field('tempholdmins') == 0: #check hold temp not applied
            return self.set_field('tempholdmins', 0)
        else:
            logging.warn("%i address, temp hold applied so won't remove set temp"%(self.set_address))

    def hold_temp(self, minutes, temp):
        """sets the temperature demand overrding the program for a set time."""
        #Believe it then returns to program.
        self.set_field('setroomtemp', temp)
        return self.set_field('tempholdmins', minutes)
        #didn't stay on if did minutes followed by temp.
        
    def release_hold_temp(self):
        """release setTemp or holdTemp back to the program."""
        return self.set_field('tempholdmins', 0)
        
    def set_holiday(self, hours):
        """sets holiday up for a defined number of hours."""
        return self.set_field('holidayhours', hours)
    
    def release_holiday(self):
        """cancels holiday mode"""
        return self.set_field('holidayhours', 0)

class ThermoStatDay(ThermoStatWeek):
    """Device class for thermostats operating daily programmode
    Heatmiser prt_e_model."""
    
    def _buildfields(self):
        """add to list of fields"""
        super(ThermoStatDay, self)._buildfields()
        self.fields.extend([
            HeatmiserFieldHeat('mon_heat', 103, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('tues_heat', 115, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('wed_heat', 127, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('thurs_heat', 139, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('fri_heat', 151, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('sat_heat', 163, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM),
            HeatmiserFieldHeat('sun_heat', 175, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM)
        ])

        self.heat_schedule = SchedulerDayHeat()

    def _connect_observers(self):
        """connect obersers to fields"""
        super(ThermoStatDay, self)._connect_observers()
        fieldnames = ['mon_heat', 'tues_heat', 'wed_heat', 'thurs_heat', 'fri_heat', 'sat_heat', 'sun_heat']
        for fieldname in fieldnames:
            getattr(self, fieldname).add_notifable_changed(self.heat_schedule.set_raw_field)

DEVICETYPES.setdefault('prt_e_model', {'week': ThermoStatWeek, 'day': ThermoStatDay})
           


"""Heatmiser Device Classes

Broadcast to all devices on the Heatmiser network
UnknownThermoStat

Ian Horsley 2018
"""
import logging

from genericdevice import HeatmiserDevice
from devices_prt_hw import ThermoStatHotWaterDay
from fields import HeatmiserFieldUnknown, HeatmiserFieldSingleReadOnly
from decorators import ListWrapperClass, run_function_on_all
from hm_constants import DEFAULT_PROTOCOL, DEFAULT_PROG_MODE, BROADCAST_ADDR
from hm_constants import MAX_AGE_LONG
from .logging_setup import csvlist

class ThermoStatUnknown(HeatmiserDevice):
    """Device class for unknown thermostats operating unknown programmode"""

    def _configure_fields(self):
        """build dict to map field name to index, map fields tables to properties and set dcb addresses."""
        super(ThermoStatUnknown, self)._configure_fields()
        self.dcb_length = 65536 #override dcb_length to prevent readall, given unknown full length # initialised in base class.
    
    def _buildfields(self):
        """add to list of fields"""
        super(ThermoStatUnknown, self)._buildfields()
        self.fields.extend([
            HeatmiserFieldUnknown('unknown', 5, MAX_AGE_LONG, 6),  # gap allows single read
            HeatmiserFieldUnknown('unknown', 12, MAX_AGE_LONG, 4),  # gap allows single read
            HeatmiserFieldSingleReadOnly('programmode', 16, [0, 1], MAX_AGE_LONG, {'day':1, 'week':0})  #0=5/2,  1= 7day
        ])
        
    def _set_expected_field_values(self):
        """set the expected values for fields that should be fixed. Overriding prevents expected model being setup."""
        self.address.expectedvalue = self.set_address
        self.DCBlen.expectedvalue = self.dcb_length

class HeatmiserBroadcastDevice(ThermoStatHotWaterDay):
    """Broadcast device class for broadcast set functions and managing reading on all devices
    Based class with most complete field list"""
    
    #List wrapper used to provide arguement to dectorator
    _controllerlist = ListWrapperClass()

    def __init__(self, network, long_name, controllerlist=None):
        self._controllerlist.list = controllerlist
        settings = {
            'address':BROADCAST_ADDR,
            'display_order': 0,
            'long_name': long_name,
            'protocol':DEFAULT_PROTOCOL,
            'expected_model':False,
            'expected_prog_mode':DEFAULT_PROG_MODE
            }
        super(HeatmiserBroadcastDevice, self).__init__(network, settings)
    
    #run read functions on all stats
    @run_function_on_all(_controllerlist)
    def read_field(self, fieldname, maxage=None):
        logging.info("All reading %s from %i controllers"%(fieldname, len(self._controllerlist.list)))
            
    @run_function_on_all(_controllerlist)
    def read_fields(self, fieldnames, maxage=None):
        logging.info("All reading %s from %i controllers"%(csvlist(fieldnames), len(self._controllerlist.list)))
        
    @run_function_on_all(_controllerlist)
    def read_air_temp(self):
        pass
    
    @run_function_on_all(_controllerlist)
    def read_temp_state(self):
        pass
    
    @run_function_on_all(_controllerlist)
    def read_water_state(self):
        pass
    
    @run_function_on_all(_controllerlist)
    def read_air_sensor_type(self):
        pass
            
    @run_function_on_all(_controllerlist)
    def read_time(self, maxage=0):
        pass
    
    #run set functions which require a read on all stats
    @run_function_on_all(_controllerlist)
    def set_temp(self, temp):
        pass

    @run_function_on_all(_controllerlist)
    def release_temp(self):
        pass
 